use crate::{error::Result, py_any_extras::PyAnyExtras};
use prost_reflect::{DynamicMessage, MapKey, Value};
use pyo3::{
    types::{IntoPyDict, PyBytes},
    PyAny, PyObject, Python, ToPyObject,
};

pub fn merge_msg_into_pyobj(obj: &PyAny, mut msg: DynamicMessage) -> Result<()> {
    for field in msg.take_fields() {
        let field_name = field.0.name();
        let proto_meta = obj.get_proto_meta()?;
        obj.setattr(
            field_name,
            map_field_value(field_name, field.1, proto_meta)?,
        )?;
    }

    let mut buf = vec![];
    for field in msg.unknown_fields() {
        field.encode(&mut buf);
    }
    if !buf.is_empty() {
        let mut unknown_fields = obj.getattr("_unknown_fields")?.extract::<Vec<u8>>()?;
        unknown_fields.append(&mut buf);
        obj.setattr("_unknown_fields", PyBytes::new(obj.py(), &unknown_fields))?;
    }

    obj.setattr("_serialized_on_wire", true)?;

    Ok(())
}

fn map_field_value(field_name: &str, field_value: Value, proto_meta: &PyAny) -> Result<PyObject> {
    let py = proto_meta.py();
    match field_value {
        Value::Bool(x) => Ok(x.to_object(py)),
        Value::Bytes(x) => Ok(x.to_object(py)),
        Value::F32(x) => Ok(x.to_object(py)),
        Value::F64(x) => Ok(x.to_object(py)),
        Value::I32(x) => Ok(x.to_object(py)),
        Value::I64(x) => Ok(x.to_object(py)),
        Value::String(x) => Ok(x.to_object(py)),
        Value::U32(x) => Ok(x.to_object(py)),
        Value::U64(x) => Ok(x.to_object(py)),
        Value::Message(msg) => {
            let obj = proto_meta.create_instance(field_name)?;
            merge_msg_into_pyobj(obj, msg)?;
            Ok(obj.to_object(py))
        }
        Value::List(ls) => Ok(ls
            .into_iter()
            .map(|x| map_field_value(field_name, x, proto_meta))
            .collect::<Result<Vec<PyObject>>>()?
            .to_object(py)),
        Value::EnumNumber(x) => {
            let cls = proto_meta.get_class(field_name)?;
            Ok(cls.call1((x,))?.to_object(py))
        }
        Value::Map(map) => {
            let res: Result<Vec<_>> = map
                .into_iter()
                .map(|(k, v)| {
                    let key = map_key(k, py);
                    let val = map_field_value(field_name, v, proto_meta)?;
                    Ok((key, val))
                })
                .collect();
            Ok(res?.into_py_dict(py).to_object(py))
        }
    }
}

fn map_key(key: MapKey, py: Python) -> PyObject {
    match key {
        MapKey::Bool(x) => x.to_object(py),
        MapKey::I32(x) => x.to_object(py),
        MapKey::I64(x) => x.to_object(py),
        MapKey::U32(x) => x.to_object(py),
        MapKey::U64(x) => x.to_object(py),
        MapKey::String(x) => x.to_object(py),
    }
}
